from os import path, getenv
from packaging.version import Version

from lxml import etree
from lxml import html

from docutils import nodes
from sphinx.highlighting import PygmentsBridge
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util import logging
from sphinx import __version__ as __sphinx_version__

from .harmonic import setup as harmonic_setup

logger = logging.getLogger(__name__)

def theme_config_setup(app):
    # Name of the repository that this doc belong to
    app.add_config_value('repository', None, 'env', [str])
    # Prefix paths with *repos/<repo>*
    app.add_config_value('monolithic', False, 'env', [bool])
    # Include sources of metadata, scripts, and styles
    app.add_config_value('core_repo', False, 'env', [bool])
    # Hide topics from the toctree expect the current one
    app.add_config_value('filter_toctree', None, 'env', [bool])
    # Setup meta tag with target depth
    app.add_config_value('target_depth', None, 'env', [str])

    app.connect("config-inited", config_inited)
    app.connect("builder-inited", builder_inited)
    app.connect("build-finished", build_finished)


setup = [
    harmonic_setup,
    theme_config_setup
]

names = ['cosmic', 'harmonic']


def config_inited(app, config):
    """
    Overwrite theme options with enviroment variables.
    Config values have higher precedance.
    Meant for value injection during CI.
    """
    if config.filter_toctree is None:
        ftoc = getenv("ADOC_FILTER_TOCTREE", default="0")
        config.filter_toctree = True if ftoc == "1" else False
    else:
        config.filter_toctree = True if config.filter_toctree else False

    # DEPRECATED
    if config.target_depth is not None or getenv("ADOC_TARGET_DEPTH", default=None) is not None:
        logger.info("ADOC_TARGET_DEPTH is deprecated and has no effect.")

def builder_inited(app):
    """
    Remove Sphinx's default provided scripts
    """
    if app.env.config.html_theme not in names:
        return

    removal = [
        "_static/documentation_options.js",
        "_static/doctools.js",
        "_static/sphinx_highlight.js"
    ]
    to_remove = []
    if app.builder.format == 'html':
        if Version(__sphinx_version__) < Version('7.2.0'):
            for js in app.builder.script_files:
                if js.filename in removal:
                    to_remove.append(js)
            for js_ in to_remove:
                app.builder.script_files.remove(js_)
        else:
            for js in app.builder._js_files:
                if js.filename in removal:
                    to_remove.append(js)
            for js_ in to_remove:
                app.builder._js_files.remove(js_)


def build_finished(app, exc):
    if app.builder.format == 'html' and not exc:
        if app.env.config.core_repo:
            from sphinx.util.fileutil import copy_asset_file
            import json

            repos = {}
            for key in app.lut['repos']:
                repos[key] = {
                    'name': app.lut['repos'][key]['name'],
                    'description': app.lut['repos'][key]['description'],
                    'category': app.lut['repos'][key]['category'],
                    'visibility': app.lut['repos'][key]['visibility'],
                    'pathname': app.lut['repos'][key]['pathname'],
                    'branch': app.lut['repos'][key]['branch']
                }
                # Miscellaneous expansions
                repos[key]['url_doc'] = app.lut['remote_doc'] + key
                repos[key]['url_source'] = app.lut['remote_https'].format(key)
                if 'topic' in app.lut['repos'][key]:
                    repos[key]['topic'] = app.lut['repos'][key]['topic']
            metadata = {'repotoc': repos}
            if app.lut['banner']['msg'] is not None:
                metadata['banner'] = app.lut['banner']

            # Extra JavaScript modules
            build_uri = path.join(app.builder.outdir, '_static')
            src_uri = path.join(path.dirname(__file__), 'harmonic', 'static')
            if app.lut['modules']['javascript'] is not None:
                if 'modules' not in metadata:
                    metadata['modules'] = {}
                metadata['modules']['javascript'] = app.lut['modules']['javascript']
                for m in app.lut['modules']['javascript']:
                    copy_asset_file(path.join(src_uri, m),
                                    path.join(build_uri, m))
            if app.lut['modules']['stylesheet'] is not None:
                if 'modules' not in metadata:
                    metadata['modules'] = {}
                metadata['modules']['stylesheet'] = app.lut['modules']['stylesheet']
                for m in app.lut['modules']['stylesheet']:
                    copy_asset_file(path.join(src_uri, m),
                                    path.join(build_uri, m))

            metadata['remote_doc'] = app.lut['remote_doc']
            metadata['remote_alt'] = app.lut['remote_alt']
            metadata['source_hostname'] = app.lut['source_hostname']
            metadata['source_hostname_raw'] = app.lut['source_hostname_raw']

            file = path.join(app.builder.outdir, 'metadata.json')
            with open(file, 'w') as f:
                json.dump(metadata, f, indent=4)


def repotoc_tree(content_root, conf_vars, pagename):
    """
    Create the empty repotoc tree linking with links of only
    the current documentation.
    Later on, with metadata.json available, the space with filled with links
    to other repos documentations.
    From the 'repository' config value, a 'current' class is added to
    the link targeting the current doc.
    """
    repo, repos = conf_vars
    root = etree.Element("root")
    home = "index.html"
    current = repo # fallback value

    repository = {}
    topics = {}
    for key in repos:
        if repos[key]['visibility'] == 'public':
            if 'topic' in repos[key]:
                for k in repos[key]['topic']:
                    topics[f"{key}/{k}"] = repos[key]['topic'][k]
            else:
                repository[key] = repos[key]['name']

    repotoc = {**topics, **repository}
    if repo is not None:
        for item in repotoc:
            attrib = {}
            if item.startswith(repo):
                if '/' in item:
                    sub = item[item.find('/')+1:]
                    href = f"{content_root}{sub}/{home}"
                    if pagename.startswith(sub):
                        attrib = {'class': 'current'}
                        current = item
                else:
                    href = f"{content_root}{home}"
                    attrib = {'class': 'current'}
                    current = repo
                link = etree.Element("a", attrib={
                    'href': href,
                    **attrib
                })
                link.text = repotoc[item]
                root.append(link)

    return (etree.tostring(root, pretty_print=True, encoding='unicode'),
            current)


def navigation_tree(app, toctree_html, content_root, pagename):
    """
    Add collapsible sections to the navigation tree.
    Adapted from
    https://github.com/pradyunsg/furo/blob/main/src/furo/navigation.py
    Parse html similar to directive/hdl_parser.py:parse_hdl_component.
    Uses checkbox~label trick, similar to
    directive/common.py:directive_base:collapsible.
    Add elements, similar to toos/hdl_render.py:hdl_component.py
    """

    conf_vars = (
        app.env.config.repository,
        app.lut['repos'],
    )

    lvl = [0]

    def get_topic(pagename):
        i = pagename.find('/')
        return pagename[0:i] if i != -1 else ''

    def filter_toctree(root, pagename):
        """
        Filter-out non-current topics/"toctrees-titles".
        Non-titled toctrees are squashed with the last toctree, e.g.
         toctrees                          |  visible
         1, 2, 3 (current)                 -> 1, 2, 3
         1 (current), 2 Info, 3, 4 User, 5 -> 1
         1, 2 Info, 3, 4 User (current), 5 -> 4, 5
         1, 2 Info, 3, 4 User, 5 (current) -> 4, 5
         1, 2 Info, 3, 4 User, 5           -> 1, 2, 3, 4, 5

        It is done like this because one reason to have multiple toctrees
        is to tweak the max depth option depending on the content,
        but still on the same topic.
        Only the System Level Documentation should have toctrees with captions.
        """
        body = root.find('./body')

        found = False
        #      Current, Elements
        tocs = [[False, []]]
        for e in body.getchildren():
            if e.tag == 'ul':
                if e.get("class") == "current":
                    tocs[-1][0] = True
                    found = True
            elif e.tag == 'p':
                tocs.append([False,[]])
            tocs[-1][1].append(e)

        # If page not on toctree, do not filter
        # e.g. /index.html, /search.html, orphan
        if not found:
            return

        for t in tocs:
            if not t[0]:
                for e in t[1]:
                    body.remove(e)

    def iterate(elem):
        for ul in elem.findall('./ul'):
            lvl[-1] += 1
            tag = '-'.join(map(str, lvl))
            _input = html.Element('input', attrib={
                'class': 'toctree-collapse',
                'type': 'checkbox',
                'name': f"toctree-collapse-{tag}",
                'id': f"toctree-collapse-{tag}"
            })
            _class = elem.get("class")
            if _class is not None and 'current' in _class:
                _input.set('checked')
            label = etree.Element("label", attrib={
                'for': f"toctree-collapse-{tag}"
            })
            label.text = ''

            for a in elem.findall('./a'):
                elem.remove(a)

            div = etree.Element("div", attrib={
                'class': 'collapse'
            })
            icon = etree.Element("div", attrib={
                'class': 'icon'
            })
            icon.text = ''
            label.append(icon)
            div.append(a)
            div.append(label)
            elem.insert(0, div)
            elem.insert(0, _input)
            lvl.append(0)
            for li in ul.findall('./li[@class]'):
                iterate(li)
            lvl.pop()

    if toctree_html:
        parser = etree.HTMLParser()
        root = etree.fromstring(toctree_html, parser)

        if app.env.config.filter_toctree:
            filter_toctree(root, pagename)
        for ul in root.findall('./body/ul'):
            for li in ul.findall('./li[@class]'):
                iterate(li)

        toctree_html = etree.tostring(root, pretty_print=True, encoding='unicode')

    _repotoc_tree, _current = repotoc_tree(content_root, conf_vars, pagename)
    if conf_vars[0] in conf_vars[1]:
        name = conf_vars[1][conf_vars[0]]['name']
    else:
        # If repository entry is not in the lut.py, use the project entry
        name = app.env.config.project
    return (toctree_html, _repotoc_tree, name, _current)


def get_pygments_theme(app):
    if app.config.pygments_style is None:
        app.config.pygments_style = 'friendly'
    light_ = app.config.pygments_style
    app.env.pygments_theme = {
        "light": PygmentsBridge("html", light_).formatter_args["style"],
        "dark": PygmentsBridge("html", "native").formatter_args["style"],
    }


def write_pygments_css(app):
    light_formatter = PygmentsBridge.html_formatter(
            style=app.env.pygments_theme["light"])
    dark_formatter = PygmentsBridge.html_formatter(
            style=app.env.pygments_theme["dark"])

    def get_styles(formatter, *, prefix):
        for line in formatter.get_linenos_style_defs():
            yield f"{prefix} {line}"
        yield from formatter.get_background_style_defs(prefix)
        yield from formatter.get_token_style_defs(prefix)

    lines = []

    lines.append("@media (prefers-color-scheme: light) {")
    lines.extend(get_styles(light_formatter,
                            prefix="body:not(.dark) .highlight"))
    lines.extend(get_styles(dark_formatter,
                            prefix="body.dark .highlight"))
    lines.append("}")

    lines.append("@media (prefers-color-scheme: dark) {")
    lines.extend(get_styles(light_formatter,
                            prefix="body.light .highlight"))
    lines.extend(get_styles(dark_formatter,
                            prefix="body:not(.light) .highlight"))
    lines.append("}")

    with open(
        path.join(app.builder.outdir, "_static", "pygments.css"),
        "w",
        encoding="utf-8",
    ) as f:
        f.write("\n".join(lines))


class wrap_elements(SphinxPostTransform):
    """A Sphinx post-transform that wraps table and div.math in a container.

    This makes it possible to handle these overflowing the content-width,
    which is necessary in a responsive theme.
    from: https://github.com/pradyunsg/furo/blob/main/src/furo/__init__.py
    """

    formats = ("html",)
    default_priority = 500

    def run(self, **kwargs) -> None:
        """Perform the post-transform on `self.document`."""
        if self.env.config.html_theme not in names:
            return

        get_nodes = (self.document.findall)
        for node in list(get_nodes(nodes.table)):
            new_node = nodes.container(classes=["table-wrapper"])
            new_node.update_all_atts(node)
            node.parent.replace(node, new_node)
            new_node.append(node)

        for node in list(get_nodes(nodes.math_block)):
            new_node = nodes.container(classes=["math-wrapper"])
            new_node.update_all_atts(node)
            node.parent.replace(node, new_node)
            new_node.append(node)



from .linkcheck_sitemap import setup as linkcheck_sitemap_setup
from .aggregate import setup as aggregate_setup
from .lfs import lfs_setup

setup = [
    linkcheck_sitemap_setup,
    aggregate_setup,
    lfs_setup,
]



"""Extend linkcheck to try links on a sitemap first.

Builds a look-up table with the entries in the sitemaps defined at linkcheck_sitemaps
config, and if the URL requested belongs to any sitemap, mark as valid, skipping the
HTTP request.
"""

from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING
from urllib.parse import urljoin, urlparse, urlunparse
from packaging.version import Version

from sphinx.util import logging
from sphinx.util import requests
from sphinx import __version__ as __sphinx_version__

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class LinkcheckSitemap:
    """Collects and stores URLs from sitemaps."""

    def __init__(self, config: Config) -> None:
        self.config = config
        self.normalized_urls: set[str] = set()

    def fetch_sitemaps(self) -> None:
        """Fetch all sitemaps and extract URLs."""
        sitemaps = getattr(self.config, 'linkcheck_sitemaps', [])

        if not sitemaps:
            return

        for url in sitemaps:
            try:
                logger.info(f'Fetching sitemap: {url}')
                self._fetch_sitemap(url)
            except Exception as exc:
                logger.warning(f'Failed to fetch sitemap {url}: {exc}')

    def _fetch_sitemap(self, url: str) -> None:
        """Fetch a single sitemap and extract URLs."""
        try:
            request_headers = getattr(self.config, 'linkcheck_request_headers', {})
            _url = urlunparse(
                urlparse(url)._replace(path='/', params='', query='', fragment='')
            )
            if _url in request_headers:
                headers = request_headers['*']
            elif '*' in request_headers:
                headers = request_headers['*']
            else:
                headers = {}

            response = requests.get(url, timeout=60, headers=headers)
            response.raise_for_status()
            content = response.content

            if url.endswith(".gz") or response.headers.get('content-type', '').endswith('gzip'):
                from io import BytesIO
                import gzip

                with gzip.GzipFile(fileobj=BytesIO(content)) as f:
                    content = f.read().decode("utf-8")

            root = ET.fromstring(content)

            if root.tag.endswith('sitemapindex'):
                self._process_sitemap_index(root, url)
            elif root.tag.endswith('urlset'):
                self._process_urlset(root)
            else:
                logger.warning(f'Unknown sitemap format in {url}')

        except ET.ParseError as exc:
            logger.warning(f'Failed to parse sitemap XML {url}: {exc}')
        except Exception as exc:
            logger.warning(f'Error fetching sitemap {url}: {exc}')

    def _process_sitemap_index(self, root: ET.Element, base_url: str) -> None:
        """Process a sitemap index file that contains references to other sitemaps."""
        namespace = self._get_namespace(root)

        for sitemap in root.findall(f'.//{{{namespace}}}sitemap'):
            loc_elem = sitemap.find(f'{{{namespace}}}loc')
            if loc_elem is not None and loc_elem.text:
                nested_sitemap_url = urljoin(base_url, loc_elem.text.strip())
                logger.debug(f'Found nested sitemap: {nested_sitemap_url}')
                self._fetch_sitemap(nested_sitemap_url)

    def _process_urlset(self, root: ET.Element) -> None:
        """Process a urlset and extract URLs."""
        namespace = self._get_namespace(root)

        for url in root.findall(f'.//{{{namespace}}}url'):
            loc_elem = url.find(f'{{{namespace}}}loc')
            if loc_elem is not None and loc_elem.text:
                self.normalized_urls.add(
                    self._normalize_url(loc_elem.text.strip())
                )

    def _get_namespace(self, root: ET.Element) -> str:
        """Extract the namespace from the XML root element."""
        default_ns = 'http://www.sitemaps.org/schemas/sitemap/0.9'

        if root.tag.startswith('{'):
            end = root.tag.find('}')
            if end > 0:
                return root.tag[1:end]

        return default_ns

    def _normalize_url(self, url: str) -> str:
        """Normalize URL for better matching (remove trailing slash, etc.)."""
        return urlunparse(
            urlparse(url)._replace(params='', query='', fragment='')
        )

    def includes(self, url: str) -> bool:
        """Check if a URL exists in the collected sitemaps."""
        normalized = self._normalize_url(url)
        if normalized in self.normalized_urls:
            return True

        return False


def builder_inited_linkcheck_sitemap(app: Sphinx) -> None:
    """Initialize the sitemap collector and fetch sitemap URLs."""
    if not hasattr(app, 'builder') or app.builder.name != 'linkcheck':
        return

    collector = LinkcheckSitemap(app.config)
    collector.fetch_sitemaps()

    app.builder.linkcheck_sitemap = collector


def process_uri_with_sitemap(app: Sphinx, uri: str) -> str | None:
    """Process URI through sitemap lookup before linkcheck validation.

    If the URI is found in any sitemap, mark as success by adding a marker to
    the URI.
    """
    if app.builder.name != 'linkcheck':
        return None

    linkcheck_sitemap = getattr(app.builder, 'linkcheck_sitemap', None)
    if not linkcheck_sitemap:
        return None

    if linkcheck_sitemap.includes(uri):
        return f'{uri} (from sitemap)'

    return None


def patch_linkcheck_worker(app: Sphinx, config: Config) -> None:
    """Patch the HyperlinkAvailabilityCheckWorker to handle in-sitemap URIs."""
    if not hasattr(app, 'builder') or app.builder.name != 'linkcheck':
        return

    from sphinx.builders.linkcheck import HyperlinkAvailabilityCheckWorker

    if Version(__sphinx_version__) > Version('8.1.3'):
        from sphinx.builders.linkcheck import _Status
        working = _Status.WORKING
    else:
        working = 'working'

    original_check = HyperlinkAvailabilityCheckWorker._check

    def patched_check(self, docname: str, uri: str, hyperlink) -> tuple:
        """Patched _check method that handles in-sitemap URIs."""
        if uri.endswith(' (from sitemap)'):
            return working, '', 0
        return original_check(self, docname, uri, hyperlink)

    HyperlinkAvailabilityCheckWorker._check = patched_check


def setup(app: Sphinx) -> ExtensionMetadata:
    """Set up the sphinx_sitemap_linkcheck extension."""

    app.add_config_value('linkcheck_sitemaps', [], 'html', types=frozenset({list, tuple}))

    app.connect('builder-inited', builder_inited_linkcheck_sitemap)
    app.connect('builder-inited', lambda app: patch_linkcheck_worker(app, app.config))

    app.connect('linkcheck-process-uri', process_uri_with_sitemap)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


from typing import TypedDict, Tuple, Dict, List, Optional


class Banner(TypedDict):
    msg: Optional[str]  # None to disable banner
    a_href: str
    a_text: str


class Modules(TypedDict):
    javascript: Optional[List[str]]
    stylesheet: Optional[List[str]]


class Repo(TypedDict):
    pathname: str
    name: str
    longname: Optional[str]
    description: str
    category: str
    branch: str
    visibility: str
    topic: Optional[Dict[str, str]]
    extra: Optional[Tuple[str, List[str], bool]]
    parent: Optional[str]


class LUT(TypedDict):
    remote_ssh: str
    remote_https: str
    remote_doc: str
    remote_alt: str
    source_hostname: str
    source_hostname_raw: str
    repos: Dict[str, Repo]
    banner: Banner
    modules: Modules


remote_ssh = "git@github.com:analogdevicesinc/{}.git"
remote_https = "https://github.com/analogdevicesinc/{}.git"
remote_doc = "https://analogdevicesinc.github.io/"
remote_alt = "https://developer.analog.com/docs/"
source_hostname = "https://github.com/analogdevicesinc/{repository}/tree/{branch}/{pathname}"
source_hostname_raw = "https://raw.githubusercontent.com/analogdevicesinc/{repository}/refs/heads/{branch}/{pathname}"

repos = {
    'documentation': Repo(
        pathname='docs',
        name='System Level',
        description='User guides and tutorials for evaluation boards, software and hardware.',
        category='system-level',
        branch='main',
        visibility='public'
    ),
    'hdl': Repo(
        pathname='docs',
        extra=(
            # cwd      # cmd            # no_parallel
            "library", ["make", "all"], False
        ),
        name='HDL',
        description='HDL libraries and projects for reference designs and prototyping systems.',
        category='hdl',
        branch='main',
        visibility='public'
    ),
    'testbenches': Repo(
        pathname='docs',
        name='HDL Testbenches',
        description='HDL testbenches for reference designs and prototyping systems.',
        category='hdl',
        branch='main',
        visibility='public',
        parent='hdl'
    ),
    'linux': Repo(
        pathname='docs',
        name='Linux',
        description='Reduced Linux kernel documentation for staging or non-upstream drivers.',
        category='driver',
        branch='main',
        visibility='public'
    ),
    'adi-kuiper-gen': Repo(
        pathname='docs',
        name='Kuiper',
        description='A Debian-based Linux distribution designed for our hardware and evaluation boards.',
        category='os',
        branch='main',
        visibility='public'
    ),
    'pyadi-iio': Repo(
        pathname='doc/source',
        name='pyadi-iio',
        longname='Python interfaces for hardware with Industrial I/O drivers',
        description='Python abstraction layer for  IIO drivers.',
        category='library',
        branch='main',
        visibility='public'
    ),
    'pyadi-dt': Repo(
        pathname='doc/source',
        name='pyadi-dt',
        longname='Device tree management tools for hardware',
        description='Library and CLI for managing device trees components.',
        category='library',
        branch='main',
        visibility='public'
    ),
    'pyadi-jif': Repo(
        pathname='doc/source',
        name='pyadi-jif',
        longname='Python Configurator for ADI JESD204 Interface Framework (JIF)',
        description='Python interface for the ADI JESD Interface Framework.',
        category='tool',
        branch='main',
        visibility='public'
    ),
    'libiio': Repo(
        pathname='doc',
        name='libiio',
        description='Cross-platform library to access IIO devices.',
        category='library',
        branch='main',
        visibility='public'
    ),
    'no-OS': Repo(
        pathname='doc/sphinx/source',
        name='no-OS',
        description='Bare-metal software framework for drivers and applications.',
        category='driver',
        branch='main',
        visibility='public'
    ),
    'precision-converters-firmware': Repo(
        pathname='doc/sphinx',
        name='Precision Converters Firmware',
        description='Embedded firmware applications for precision converters on SDP-K1 and others.',
        category='library',
        branch='main',
        visibility='public'
    ),
    'PrecisionToolbox': Repo(
        pathname='docs/source',
        name='Precision Toolbox',
        description='Set of tools to interface precision converters within MATLAB and Simulink.',
        category='library',
        branch='main',
        visibility='public'
    ),
    'adi_ros2': Repo(
        pathname='src/adi_ros2/doc',
        name='ROS2',
        description='Robotics SDK with ROS 2 packages and tools in pre-configured container images.',
        category='toolkit',
        branch='humble',
        visibility='public'
    ),
    'scopy': Repo(
        pathname='docs',
        name='Scopy',
        description='Multi-functional software toolset with strong capabilities for signal analysis.',
        category='tool',
        branch='main',
        visibility='public'
    ),
    'doctools': Repo(
        pathname='docs',
        name='Doctools',
        description='Documentation guidelines, extensions, and tools.',
        category='tool',
        branch='main',
        visibility='public'
    ),
}

banner = Banner(
    msg=None,
    a_href='',
    a_text=''
)

"""
Allows to inject extra scripts and stylesheets on the hosted documentation.
The app.js looks the metadata.json and insert tags to load each module of the array.
All files must be (generated) at the theme static folder, to be copied over the
documentation build.
The advantage of this approach is that regardless of the tools' built doc version,
it will always fetch the latest and greatest.
See also: ci/rollup.config.app.mjs

To provide a slight speed boost, the JavaScript 'bets' that the metadata.json
contains extra.*.*, appending to the page before the metadata loads,
other items will be appended after metadata.json is fetched.
"""
modules = Modules(
    javascript=['extra.umd.js'],
    stylesheet=['extra.min.css']
)


def get_lut():
    # TODO dynamic lut fetch
    return LUT(remote_ssh=remote_ssh,
               remote_https=remote_https,
               remote_doc=remote_doc,
               remote_alt=remote_alt,
               source_hostname=source_hostname,
               source_hostname_raw=source_hostname_raw,
               repos=repos,
               banner=banner,
               modules=modules)



from os import environ

from sphinx.transforms.post_transforms import SphinxTransform
from sphinx.util import logging, docname_join
from sphinx import addnodes

logger = logging.getLogger(__name__)


class namespace_pending_xfer(SphinxTransform):
    """
    Patch pending xfer to have the format with local with higher precedence,
    if doesn't exist, try external:

        :ref:`hdl:user_guide`
    Rules:

    * Ref:
      - external: Patch all, to always try local first.
      - local: only if root path matches inventory list, creating "namespaces"
        per repository.

    * Doc:
      - external: Check if external would match current tree, if so, convert
        into local.
      - local: Check if matches current tree (only current inventory, if not,
        convert into external.

    This methodology allows to create custom documents having to patch only
    explicit references `.. _my-ref:`, while the xfer are resolved during build.
    """
    default_priority = 3

    def apply(self, **kwargs) -> None:
        inventories = self.env.config.intersphinx_mapping.keys()
        fromdocname = self.env.docname
        for node in self.document.findall(addnodes.pending_xref):
            if node['reftype'] == 'doc':
                refdoc = node.get('refdoc', fromdocname)
                if 'intersphinx' in node:
                    docname = node['inventory'] + '/' + node['reftarget'].strip('/')
                    if docname in self.env.found_docs:
                        # all_docs is filled at each doc read, after the Transforms,
                        # use found_docs instead.
                        node['reftarget'] = '/' + docname
                        del node['intersphinx']
                        del node['inventory']
                else:
                    docname = docname_join(refdoc, node['reftarget'])
                    if docname not in self.env.found_docs:
                        node['intersphinx'] = True
                        reftarget = docname.split('/')
                        node['inventory'] = reftarget.pop(0)
                        node['reftarget'] = '/'.join(reftarget)
            else:
                if 'intersphinx' in node:
                    node['reftarget'] = node['inventory'] + ":" + node['reftarget']
                    del node['intersphinx']
                    del node['inventory']
                else:
                    inventory = node.get('refdoc', fromdocname).split('/')[0]
                    match = next((k for k in inventories if (inventory == k or
                                                             inventory.startswith(k + '_'))),
                                 None)

                    if match and node['reftarget'].find(':') == -1:
                        node['reftarget'] = match + ":" + node['reftarget']


def setup(app):
    if "ADOC_CUSTOM_DOC" in environ:
        app.add_transform(namespace_pending_xfer)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import click

from .serve import serve
from .hdl_render import hdl_render
from .hdl_gen import hdl_gen
from .aggregate import aggregate
from .custom_doc import custom_doc
from .search import search


class AliasGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        aliases = {
            "author-mode": "serve",
            "server": "serve",
        }
        if cmd_name in aliases:
            cmd_name_ = aliases[cmd_name]
            click.echo("Warning: redirected command"
                       f"'{cmd_name}' to '{cmd_name_}'", err=True)
            cmd_name = cmd_name_

        cmd_name = aliases.get(cmd_name, cmd_name)
        return super().get_command(ctx, cmd_name)


@click.group(cls=AliasGroup)
def entry_point():
    """
    CLI to provide entry points to adi_doctools methods.
    """
    pass


commands = [
    serve,
    hdl_render,
    hdl_gen,
    aggregate,
    custom_doc,
    search,
]

for cmd in commands:
    entry_point.add_command(cmd)
